import copy

import numpy as np
from typing import cast

import autofit as af
from autolens import exc
from autolens.lens import lens_data as ld, lens_fit
from autolens.model.galaxy import galaxy as g
from autolens.model.hyper import hyper_data as hd
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline.phase.phase import setup_phase_mask
from autolens.pipeline.plotters import hyper_plotters
from .hyper_phase import HyperPhase


class HyperGalaxyPhase(HyperPhase):
    def __init__(
        self, phase, include_sky_background=False, include_noise_background=False
    ):
        super().__init__(phase=phase, hyper_name="hyper_galaxies")
        self.include_sky_background = include_sky_background
        self.include_noise_background = include_noise_background

    class Analysis(af.Analysis):
        def __init__(
            self, lens_data, hyper_model_image_1d, hyper_galaxy_image_1d_path_dict
        ):
            """
            An analysis to fit the noise for a single galaxy image.
            Parameters
            ----------
            lens_data: LensData
                lens data, including an image and noise
            hyper_model_image_1d: ndarray
                An image produce of the overall system by a model
            hyper_galaxy_image_1d_path_dict: ndarray
                The contribution of one galaxy to the model image
            """

            self.lens_data = lens_data

            self.hyper_model_image_1d = hyper_model_image_1d
            self.hyper_galaxy_image_1d_path_dict = hyper_galaxy_image_1d_path_dict

            self.plot_hyper_galaxy_subplot = af.conf.instance.visualize.get(
                "plots", "plot_hyper_galaxy_subplot", bool
            )

        def visualize(self, instance, image_path, during_analysis):

            if self.plot_hyper_galaxy_subplot:

                hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)

                hyper_background_noise = self.hyper_background_noise_for_instance(
                    instance=instance
                )

                hyper_model_image_2d = self.lens_data.grid.scaled_array_2d_from_array_1d(
                    array_1d=self.hyper_model_image_1d
                )

                for galaxy_path, galaxy in instance.path_instance_tuples_for_class(
                    g.Galaxy
                ):

                    if galaxy_path in self.hyper_galaxy_image_1d_path_dict:
                        hyper_galaxy_image_2d = self.lens_data.grid.scaled_array_2d_from_array_1d(
                            array_1d=self.hyper_galaxy_image_1d_path_dict[galaxy_path]
                        )

                        contribution_map_2d = galaxy.hyper_galaxy.contribution_map_from_hyper_images(
                            hyper_model_image=hyper_model_image_2d,
                            hyper_galaxy_image=hyper_galaxy_image_2d,
                        )

                        fit_normal = lens_fit.LensDataFit(
                            image_1d=self.lens_data.image_1d,
                            noise_map_1d=self.lens_data.noise_map_1d,
                            mask_1d=self.lens_data.mask_1d,
                            model_image_1d=self.hyper_model_image_1d,
                            grid=self.lens_data.grid,
                        )

                        fit = self.fit_for_hyper_galaxy(
                            hyper_galaxy=galaxy.hyper_galaxy,
                            hyper_image_sky=hyper_image_sky,
                            hyper_background_noise=hyper_background_noise,
                        )

                        hyper_plotters.plot_hyper_galaxy_subplot(
                            hyper_galaxy_image=hyper_galaxy_image_2d,
                            contribution_map=contribution_map_2d,
                            noise_map=self.lens_data.noise_map(return_in_2d=True),
                            hyper_noise_map=fit.noise_map(return_in_2d=True),
                            chi_squared_map=fit_normal.chi_squared_map(
                                return_in_2d=True
                            ),
                            hyper_chi_squared_map=fit.chi_squared_map(
                                return_in_2d=True
                            ),
                            output_path=image_path,
                            output_format="png",
                        )

        def fit(self, instance):
            """
            Fit the model image to the real image by scaling the hyper_galaxies noise.
            Parameters
            ----------
            instance: ModelInstance
                A model instance with a hyper_galaxies galaxy property
            Returns
            -------
            fit: float
            """

            hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)

            hyper_background_noise = self.hyper_background_noise_for_instance(
                instance=instance
            )

            fit = self.fit_for_instance(instance=instance)

            return fit.figure_of_merit

        @staticmethod
        def hyper_image_sky_for_instance(instance):
            if hasattr(instance, "hyper_image_sky"):
                return instance.hyper_image_sky

        @staticmethod
        def hyper_background_noise_for_instance(instance):
            if hasattr(instance, "hyper_background_noise"):
                return instance.hyper_background_noise

        def fit_for_hyper_galaxy(
            self, hyper_galaxy, hyper_image_sky, hyper_background_noise
        ):

            if hyper_image_sky is not None:
                image_1d = hyper_image_sky.image_scaled_sky_from_image(
                    image=self.lens_data.image_1d
                )
            else:
                image_1d = self.lens_data.image_1d

            if hyper_background_noise is not None:
                noise_map_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
                    noise_map=self.lens_data.noise_map_1d
                )
            else:
                noise_map_1d = self.lens_data.noise_map_1d

            hyper_noise_1d = hyper_galaxy.hyper_noise_map_from_hyper_images_and_noise_map(
                hyper_model_image=self.hyper_model_image_1d,
                hyper_galaxy_image=self.hyper_galaxy_image_1d_path_dict,
                noise_map=self.lens_data.noise_map_1d,
            )

            hyper_noise_map_1d = noise_map_1d + hyper_noise_1d

            return lens_fit.LensDataFit(
                image_1d=image_1d,
                noise_map_1d=hyper_noise_map_1d,
                mask_1d=self.lens_data.mask_1d,
                model_image_1d=self.hyper_model_image_1d,
                grid=self.lens_data.grid,
            )

        def fit_for_instance(self, instance):

            if instance.hyper_image_sky is not None:
                image_1d = instance.hyper_image_sky.image_scaled_sky_from_image(
                    image=self.lens_data.image_1d
                )
            else:
                image_1d = self.lens_data.image_1d

            if instance.hyper_background_noise is not None:
                noise_map_1d = instance.hyper_background_noise.noise_map_scaled_noise_from_noise_map(
                    noise_map=self.lens_data.noise_map_1d
                )
            else:
                noise_map_1d = self.lens_data.noise_map_1d

            for galaxy_path, galaxy in instance.path_instance_tuples_for_class(
                g.Galaxy
            ):

                hyper_noise_1d = np.zeros(shape=self.hyper_model_image_1d.shape)

                if galaxy_path in self.hyper_galaxy_image_1d_path_dict:

                    print(galaxy.hyper_galaxy)

                    if galaxy.hyper_galaxy is not None:
                        hyper_galaxy_image_1d = self.hyper_galaxy_image_1d_path_dict[
                            galaxy_path
                        ]

                        hyper_noise_1d += galaxy.hyper_galaxy.hyper_noise_map_from_hyper_images_and_noise_map(
                            hyper_model_image=self.hyper_model_image_1d,
                            hyper_galaxy_image=hyper_galaxy_image_1d,
                            noise_map=self.lens_data.noise_map_1d,
                        )

            hyper_noise_map_1d = noise_map_1d + hyper_noise_1d

            return lens_fit.LensDataFit(
                image_1d=image_1d,
                noise_map_1d=hyper_noise_map_1d,
                mask_1d=self.lens_data.mask_1d,
                model_image_1d=self.hyper_model_image_1d,
                grid=self.lens_data.grid,
            )

        @classmethod
        def describe(cls, instance):
            return "Running hyper_galaxies galaxy fit for HyperGalaxy:\n{}".format(
                instance.hyper_galaxy
            )

    def run_hyper(self, data, results=None):
        """
        Run a fit for each galaxy from the previous phase.
        Parameters
        ----------
        data: LensData
        results: ResultsCollection
            Results from all previous phases
        mask: Mask
            The mask
        positions
        Returns
        -------
        results: HyperGalaxyResults
            A collection of results, with one item per a galaxy
        """
        phase = self.make_hyper_phase()

        lens_data = ld.LensData(
            ccd_data=data,
            mask=results.last.mask_2d,
            sub_grid_size=cast(al.PhaseImaging, phase).sub_grid_size,
            trimmed_psf_shape=cast(al.PhaseImaging, phase).psf_shape,
            positions=results.last.positions,
            pixel_scale_interpolation_grid=cast(
                al.PhaseImaging, phase
            ).pixel_scale_interpolation_grid,
            pixel_scale_binned_cluster_grid=cast(
                al.PhaseImaging, phase
            ).pixel_scale_binned_cluster_grid,
            inversion_pixel_limit=cast(al.PhaseImaging, phase).inversion_pixel_limit,
            uses_cluster_inversion=cast(al.PhaseImaging, phase).uses_cluster_inversion,
        )

        hyper_model_image_1d = results.last.hyper_model_image_1d
        hyper_galaxy_image_1d_path_dict = results.last.hyper_galaxy_image_1d_path_dict

        hyper_result = copy.deepcopy(results.last)
        hyper_result.variable = hyper_result.variable.copy_with_fixed_priors(
            hyper_result.constant
        )

        hyper_result.analysis.hyper_model_image_1d = hyper_model_image_1d
        hyper_result.analysis.hyper_galaxy_image_1d_path_dict = (
            hyper_galaxy_image_1d_path_dict
        )

        for path, galaxy in results.last.path_galaxy_tuples:

            optimizer = phase.optimizer.copy_with_name_extension(extension=path[-1])

            optimizer.phase_tag = ""

            # TODO : This is a HACK :O

            if hasattr(results.last.constant, "galaxies"):
                optimizer.variable.galaxies = results.last.constant.galaxies

            if hasattr(results.last.constant, "galaxies"):
                optimizer.variable.galaxies = results.last.constant.galaxies

            if hasattr(results.last.constant, "galaxies"):
                optimizer.variable.galaxies = results.last.constant.galaxies

            optimizer.const_efficiency_mode = af.conf.instance.non_linear.get(
                "MultiNest", "extension_hyper_galaxy_const_efficiency_mode", bool
            )
            optimizer.sampling_efficiency = af.conf.instance.non_linear.get(
                "MultiNest", "extension_hyper_galaxy_sampling_efficiency", float
            )
            optimizer.n_live_points = af.conf.instance.non_linear.get(
                "MultiNest", "extension_hyper_galaxy_n_live_points", int
            )

            # TODO : WE need to associate each galaxy, for each optimizer, here, to get the new hyper_galaxies galaxy phase to work.

            optimizer.variable.hyper_galaxy = g.HyperGalaxy

            if self.include_sky_background:
                optimizer.variable.hyper_image_sky = hd.HyperImageSky

            if self.include_noise_background:
                optimizer.variable.hyper_background_noise = hd.HyperBackgroundNoise

            # If array is all zeros, galaxy did not have image in previous phase and
            # should be ignored
            if not np.all(hyper_galaxy_image_1d_path_dict[path] == 0):

                analysis = self.Analysis(
                    lens_data=lens_data,
                    hyper_model_image_1d=hyper_model_image_1d,
                    hyper_galaxy_image_1d_path_dict=hyper_galaxy_image_1d_path_dict,
                )

                result = optimizer.fit(analysis)

                def transfer_field(name):
                    if hasattr(result.constant, name):
                        setattr(
                            hyper_result.constant.object_for_path(path),
                            name,
                            getattr(result.constant, name),
                        )
                        setattr(
                            hyper_result.variable.object_for_path(path),
                            name,
                            getattr(result.variable, name),
                        )

                transfer_field("hyper_galaxies")
                transfer_field("hyper_image_sky")
                transfer_field("hyper_background_noise")

        return hyper_result



import autofit as af
import autolens as al

import os

# This tool allows one to input a set of positions of a multiply imaged strongly lensed source, corresponding to a set
# positions / pixels which are anticipated to trace to the same location in the source-plane.

# A non-linear sampler uses these positions to discard the mass-models where they do not trace within a threshold of
# one another, speeding up the analysis and removing unwanted solutions with too much / too little mass.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# The 'data name' is the name of the data folder and 'data_name' the folder the positions are stored in e.g,
# the positions will be output as '/workspace/data/data_type/data_name/positions.dat'.
data_type = "example"
data_name = "lens_sersic_sie__source_sersic"

# Create the path where the mask will be output, which in this case is
# '/workspace/data/example/lens_light_and_x1_source/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

# If you use this tool for your own instrument, you *must* double check this pixel scale is correct!
pixel_scale = 0.1

# First, load the CCD imaging data, so that the positions can be plotted over the strong lens image.
image = al.load_image(
    image_path=data_path + "image.fits", image_hdu=0, pixel_scale=pixel_scale
)

# Now, create a set of positions, which is simply a python list of (y,x) values.
positions = [[[0.8, 1.45], [1.78, -0.4], [-0.95, 1.38], [-0.83, -1.04]]]

# These are the positions for the example lens 'lens_sersic_sie__source_sersic__2'
# positions = [[[0.44, 0.60], [-1.4, -0.41], [0.15, -1.45]]]

# These are the positions for the example lens 'lens_sie__source_sersic__2'
# positions = [[[1.28, -1.35], [-0.5, 0.7]]]

# These are the positions for the example lens 'lens_mass__source_sersic_x2'
# positions = [[[2.16, -1.3], [-0.65, 0.45]]]

# We can infact input multiple lists of positions (commented out below), which corresponds to pixels which are \
# anticipated to map to different multiply imaged regions of the source-plane (e.g. you would need something like \
# spectra to be able to do this)
# Images of source 1           # Images of source 2
# positions = [[[1.0, 1.0], [2.0, 0.5]], [[-1.0, -0.1], [2.0, 2.0], [3.0, 3.0]]]

# Now lets plot the image and positions, so we can check that the positions overlap different regions of the source.
al.data_plotters.plot_image(image=image, positions=positions)

# Now we're happy with the positions, lets output them to the data folder of the lens, so that we can load them from a
# .dat file in our pipelines!
al.output_positions(positions=positions, positions_path=data_path + "positions.dat")


# These commented out lines would create the positions for the example_multi_plane instrument.
# lens_name = 'example_multi_plane'
# pixel_scale = 0.05
# positions = [[[0.8, 1.12], [-0.64, 1.13], [1.38, -0.2], [-0.72, -0.83]]]



import autofit as af
import autolens as al

import os

from workspace.tools.data_making.samples import tools

#### UNFINISHE CODE ####

# This tool creates a sample of strong lenses with no lens light component. These lenses use a SIE mass profile and a
# source galaxy generated with x1 Sersic profile. This sample is used to test pipelines and runners.

# The 'data name' is the name of the data folder and 'data_name' the folder the data is stored in, e.g:

# The image will be output as '/workspace/data/data_type/data_name/image.fits'.
# The noise-map will be output as '/workspace/data/data_type/data_name/lens_name/noise_map.fits'.
# The psf will be output as '/workspace/data/data_type/data_name/psf.fits'.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../../../".format(os.path.dirname(os.path.realpath(__file__)))

# (these files are already in the workspace and are remade running this script)
data_imaging_type = "Euclid"
data_lens_light = "no_lens_light"
data_lens_type = "sie_shear_x1_source"

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path,
    folder_names=["data", "samples", data_lens_light, data_lens_type],
)

# The pixel scale of the image to be simulated
ccd_shape = (120, 120)
pixel_scale = 0.1
sub_grid_size = 1
plot_ccd = True

# Simulate a simple Gaussian PSF for the image.
psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.1, pixel_scale=pixel_scale)

###### SIE + SHEAR LENS 1 #######

data_name = "sie_shear_x1_source_1"
lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.05, phi=90.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=60.0,
        intensity=0.3,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
)

tools.from_data_lens__source_galaxy_simulate_and_output_ccd_data(
    data_path=data_path,
    data_name=data_name,
    ccd_shape=ccd_shape,
    pixel_scale=pixel_scale,
    psf=psf,
    lens_galaxy=lens_galaxy,
    source_galaxy=source_galaxy,
    sub_grid_size=sub_grid_size,
    plot_ccd=plot_ccd,
)

###### SIE + SHEAR LENS 2 #######

data_name = "sie_shear_x1_source_2"
lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.2, axis_ratio=0.9, phi=105.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.1, phi=145.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.3, 0.3),
        axis_ratio=0.6,
        phi=120.0,
        intensity=0.3,
        effective_radius=2.0,
        sersic_index=1.5,
    ),
)

tools.from_data_lens__source_galaxy_simulate_and_output_ccd_data(
    data_path=data_path,
    data_name=data_name,
    ccd_shape=ccd_shape,
    pixel_scale=pixel_scale,
    psf=psf,
    lens_galaxy=lens_galaxy,
    source_galaxy=source_galaxy,
    sub_grid_size=sub_grid_size,
    plot_ccd=plot_ccd,
)

###### SIE + SHEAR LENS 3 #######

data_name = "sie_shear_x1_source_3"
lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=0.8, axis_ratio=0.9, phi=10.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.02, phi=30.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(-0.2, 0.4),
        axis_ratio=0.4,
        phi=160.0,
        intensity=0.1,
        effective_radius=0.2,
        sersic_index=2.5,
    ),
)

tools.from_data_lens__source_galaxy_simulate_and_output_ccd_data(
    data_path=data_path,
    data_name=data_name,
    ccd_shape=ccd_shape,
    pixel_scale=pixel_scale,
    psf=psf,
    lens_galaxy=lens_galaxy,
    source_galaxy=source_galaxy,
    sub_grid_size=sub_grid_size,
    plot_ccd=plot_ccd,
)

###### SIE + SHEAR LENS 4 #######

data_name = "sie_shear_x1_source_4"
lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.3, axis_ratio=0.9, phi=100.0
    ),
    shear=al.mass_profiles.ExternalShear(magnitude=0.15, phi=170.0),
)

source_galaxy = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.5, -0.1),
        axis_ratio=0.8,
        phi=60.0,
        intensity=0.2,
        effective_radius=0.5,
        sersic_index=3.0,
    ),
)

tools.from_data_lens__source_galaxy_simulate_and_output_ccd_data(
    data_path=data_path,
    data_name=data_name,
    ccd_shape=ccd_shape,
    pixel_scale=pixel_scale,
    psf=psf,
    lens_galaxy=lens_galaxy,
    source_galaxy=source_galaxy,
    sub_grid_size=sub_grid_size,
    plot_ccd=plot_ccd,
)



import autolens as al
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion.plotters import mapper_plotters

# In the previous example, we made a mapper from a rectangular pixelization. However, it wasn't clear what a mapper
# was actually mapping. Infact, it didn't do much mapping at all! Therefore, in this tutorial, we'll cover mapping.

# To begin, lets simulate and load an image - it'll be clear why we're doing this in a moment.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(150, 150), pixel_scale=0.05, sub_grid_size=2
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.6
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.7,
            phi=135.0,
            intensity=0.2,
            effective_radius=0.2,
            sersic_index=2.5,
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Lets simulate our CCD data.
ccd_data = simulate()
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# Now, lets set up our grids (using the image above).
grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, sub_grid_size=2
)

# Our tracer will use the same lens galaxy and source galaxy that we used to Simulate the CCD data (although, becuase
# we're modeling the source with a pixel-grid, we don't need to supply its light profile).
lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.6
    ),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, al.Galaxy(redshift=1.0)])

source_plane_grid = tracer.traced_grids_of_planes_from_grid(grid=grid)[1]

# Next, we setup our pixelization and mapper using the tracer's source-plane grid.
rectangular = pix.Rectangular(shape=(25, 25))

mapper = rectangular.mapper_from_grid_and_pixelization_grid(grid=source_plane_grid)

# We're going to plot our mapper alongside the image we used to generate the source-plane grid.
mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data, mapper=mapper, should_plot_grid=True
)

# The pixels in the image map to the pixels in the source-plane, and visa-versa. Lets highlight a set of
# image-pixels in both the image and source-plane.
mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data,
    mapper=mapper,
    should_plot_grid=True,
    image_pixels=[[range(0, 100)], [range(900, 1000)]],
)

# That's nice, and we can see the mappings, but it isn't really what we want to know, is it? We really want to go the
# other way, and see how our source-pixels map to the image. This is where mappers come into their own, as they let us
# map all the points in a given source-pixel back to the image. Lets map source pixel 313, the central
# source-pixel, to the image.
mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data, mapper=mapper, should_plot_grid=True, source_pixels=[[312]]
)

# And there we have it - multiple imaging in all its glory. Try changing the source-pixel indexes of the line below.
# This will give you a feel for how different regions of the source-plane map to the image.
mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data,
    mapper=mapper,
    should_plot_grid=True,
    source_pixels=[[312, 318], [412]],
)

# Okay, so I think we can agree, mappers map things! More specifically, they map our source-plane pixels to pixels in the
# observed image of a strong lens.

# Finally, lets do the same as above, but using a masked image. By applying a mask, the mapper will only map
# image-pixels inside the mask. This removes the (many) image pixels at the edge of the image, where the source
# isn't present. These pixels also pad-out the source-plane, thus by removing them our source-plane reduces in size.
# Lets just have a quick look at these edges pixels:
mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data,
    mapper=mapper,
    should_plot_grid=True,
    source_pixels=[[0, 1, 2, 3, 4, 5, 6, 7], [620, 621, 622, 623, 624]],
)

# Lets use an annular mask, which will capture the ring-like shape of the lensed source galaxy.
mask = al.Mask.circular_annular(
    shape=ccd_data.shape,
    pixel_scale=ccd_data.pixel_scale,
    inner_radius_arcsec=1.0,
    outer_radius_arcsec=2.2,
)

# Lets quickly confirm the annuli capture the source's light.
al.ccd_plotters.plot_image(ccd_data=ccd_data, mask=mask)

# As usual, we setup our image and mask up as lens data and create a tracer using its (now masked) grids.
lens_data = al.LensData(ccd_data=ccd_data, mask=mask)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, al.Galaxy(redshift=1.0)])

source_plane_grid = tracer.traced_grids_of_planes_from_grid(grid=lens_data.grid)[1]

# Finally, we use the masked source-plane grid to setup a new mapper (using the same rectangular 25 x 25
# pixelization as before).
mapper = rectangular.mapper_from_grid_and_pixelization_grid(grid=source_plane_grid)

# Lets have another look.
mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data, mask=mask, mapper=mapper, should_plot_grid=True
)

# Woah! Look how much closer we are to the source-plane (The axis sizes have decreased from ~ -2.5" -> 2.5" to
# ~ -0.6" to 0.6"). We can now really see the diamond of points in the centre of the source-plane (for those who have
# been reading up, this diamond is called the 'caustic').
mapper_plotters.plot_image_and_mapper(
    ccd_data=ccd_data,
    mask=mask,
    mapper=mapper,
    should_plot_grid=True,
    source_pixels=[[312], [314], [316], [318]],
)

# Great - tutorial 2 down! We've learnt about mappers, which map things, and we used them to understand how the image
# and source plane map to one another. Your exercises are:

# 1) Change the einstein radius of the lens galaxy in small increments (e.g. einstein radius 1.6" -> 1.55").
#    As the radius deviates from 1.6" (the input value of the simulated lens), what do you notice about where the
#    points map from the centre of the source-plane (where the source-galaxy is simulated, e.g. (0.0", 0.0"))?

# 2) Incrementally increase the axis ratio of the lens's mass profile to 1.0. What happens to quadruple imaging?

# 3) Now, finally, think - how is all of this going to help us actually model lenses? We've said we're going to
#    reconstruct our source galaxies on the pixel-grid. So, how does knowing how each pixel maps to the image
#    actually help us? If you've not got any bright ideas, then worry not - that exactly what we're going to cover
#    in the next tutorial.



import autofit as af
import autolens as al

# In this pipeline, we'll perform an initialize analysis which fits an image with a source galaxy and no lens light
# component. The pipeline is as follows:

# Phase 1:

# Description: initialize the lens mass model and source light profile.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: None


def make_pipeline(
    pipeline_settings,
    phase_folders=None,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    signal_to_noise_limit=None,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    pixel_scale_interpolation_grid=None,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_init_hyper__lens_sie__source_sersic"

    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings(
        hyper_galaxies=pipeline_settings.hyper_galaxies,
        hyper_image_sky=pipeline_settings.hyper_image_sky,
        hyper_background_noise=pipeline_settings.hyper_background_noise,
        include_shear=pipeline_settings.include_shear,
    )

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    ### SETUP SHEAR ###

    # If the pipeline should include shear, add this class below so that it enters the phase.

    # After this pipeline this shear class is passed to all subsequent pipelines, such that the shear is either
    # included or omitted throughout the entire pipeline.

    if pipeline_settings.include_shear:
        shear = al.mass_profiles.ExternalShear
    else:
        shear = None

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy (y,x) centre such that we assume the image is centred around the lens galaxy.

    phase1 = al.PhaseImaging(
        phase_name="phase_1__lens_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=shear,
            ),
            source=al.GalaxyModel(
                redshift=redshift_source, light=al.light_profiles.EllipticalSersic
            ),
        ),
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    phase1 = phase1.extend_with_multiple_hyper_phases(
        hyper_galaxy=pipeline_settings.hyper_galaxies,
        include_background_sky=pipeline_settings.hyper_image_sky,
        include_background_noise=pipeline_settings.hyper_background_noise,
    )

    return al.PipelineImaging(pipeline_name, phase1, hyper_mode=True)



import autofit as af
import matplotlib

backend = af.conf.instance.visualize.get("figures", "backend", str)
matplotlib.use(backend)
import numpy as np

from autolens.plotters import array_plotters
from autolens.plotters import grid_plotters
from autolens.plotters import line_yx_plotters


def plot_image(
    image,
    plot_origin=True,
    grid=None,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    should_plot_border=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=30,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="image",
):
    """Plot the observed image of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : ScaledSquarePixelArray
        The image of the data.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    image_plane_pix_grid : ndarray or instrument.array.grid_stacks.PixGrid
        If an adaptive pixelization whose pixels are formed by tracing pixels from the data, this plots those pixels \
        over the immage.
    """
    origin = get_origin(array=image, plot_origin=plot_origin)

    array_plotters.plot_array(
        array=image,
        origin=origin,
        grid=grid,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        should_plot_border=should_plot_border,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_noise_map(
    noise_map,
    plot_origin=True,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Noise-Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="noise_map",
):
    """Plot the noise_map of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    noise_map : ScaledSquarePixelArray
        The noise map of the data.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    """
    origin = get_origin(array=noise_map, plot_origin=plot_origin)

    array_plotters.plot_array(
        array=noise_map,
        origin=origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_psf(
    psf,
    plot_origin=True,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="PSF",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="psf",
):
    """Plot the PSF of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    signal_to_noise_map : ScaledSquarePixelArray
        The psf of the data.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    """
    origin = get_origin(array=psf, plot_origin=plot_origin)

    array_plotters.plot_array(
        array=psf,
        origin=origin,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_signal_to_noise_map(
    signal_to_noise_map,
    plot_origin=True,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Signal-To-Noise-Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="signal_to_noise_map",
):
    """Plot the signal-to-noise_map of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    signal_to_noise_map : ScaledSquarePixelArray
        The signal-to-noise map of the data.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    """
    origin = get_origin(array=signal_to_noise_map, plot_origin=plot_origin)

    array_plotters.plot_array(
        array=signal_to_noise_map,
        origin=origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_absolute_signal_to_noise_map(
    absolute_signal_to_noise_map,
    plot_origin=True,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Absolute Signal-To-Noise-Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="absolute_signal_to_noise_map",
):
    """Plot the absolute signal-to-noise map of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    absolute_signal_to_noise_map : ScaledSquarePixelArray
        The absolute signal-to-noise map of the data.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    """
    origin = get_origin(array=absolute_signal_to_noise_map, plot_origin=plot_origin)

    array_plotters.plot_array(
        array=absolute_signal_to_noise_map,
        origin=origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_potential_chi_squared_map(
    potential_chi_squared_map,
    plot_origin=True,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Potential Chi-Squared Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="potential_chi_squared_map",
):
    """Plot the signal-to-noise_map of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    potential_chi_squared_map : ScaledSquarePixelArray
        The signal-to-noise map of the data.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    """
    origin = get_origin(array=potential_chi_squared_map, plot_origin=plot_origin)

    array_plotters.plot_array(
        array=potential_chi_squared_map,
        origin=origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_visibilities(
    visibilities,
    visibilities_noise_map=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    cmap="jet",
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Visibilities",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="visibilities",
):
    """Plot the observed image of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : ScaledSquarePixelArray
        The image of the data.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    image_plane_pix_grid : ndarray or instrument.array.grid_stacks.PixGrid
        If an adaptive pixelization whose pixels are formed by tracing pixels from the data, this plots those pixels \
        over the immage.
    """

    grid_plotters.plot_grid(
        grid=visibilities,
        colors=visibilities_noise_map,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        cmap=cmap,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_u_wavelengths(
    uv_wavelengths,
    as_subplot=False,
    label="Wavelengths",
    units="",
    kpc_per_arcsec=None,
    figsize=(14, 7),
    plot_axis_type="linear",
    ylabel="U-Wavelength",
    title="U-Wavelengths",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    legend_fontsize=12,
    output_path=None,
    output_format="show",
    output_filename="u_wavelengths",
):
    """Plot the observed image of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : ScaledSquarePixelArray
        The image of the data.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    image_plane_pix_grid : ndarray or instrument.array.grid_stacks.PixGrid
        If an adaptive pixelization whose pixels are formed by tracing pixels from the data, this plots those pixels \
        over the immage.
    """

    line_yx_plotters.plot_line(
        y=uv_wavelengths[:, 0],
        x=None,
        as_subplot=as_subplot,
        label=label,
        plot_axis_type=plot_axis_type,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        title=title,
        ylabel=ylabel,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        legend_fontsize=legend_fontsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_v_wavelengths(
    uv_wavelengths,
    as_subplot=False,
    label="Wavelengths",
    units="",
    kpc_per_arcsec=None,
    figsize=(14, 7),
    plot_axis_type="linear",
    ylabel="V-Wavelength",
    title="V-Wavelengths",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    legend_fontsize=12,
    output_path=None,
    output_format="show",
    output_filename="v_wavelengths",
):
    """Plot the observed image of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : ScaledSquarePixelArray
        The image of the data.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    image_plane_pix_grid : ndarray or instrument.array.grid_stacks.PixGrid
        If an adaptive pixelization whose pixels are formed by tracing pixels from the data, this plots those pixels \
        over the immage.
    """

    line_yx_plotters.plot_line(
        y=uv_wavelengths[:, 1],
        x=None,
        as_subplot=as_subplot,
        label=label,
        plot_axis_type=plot_axis_type,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        title=title,
        ylabel=ylabel,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        legend_fontsize=legend_fontsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_primary_beam(
    primary_beam,
    plot_origin=True,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Primary Beam",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="primary_beam",
):
    """Plot the PSF of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    signal_to_noise_map : ScaledSquarePixelArray
        The primary_beam of the data.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    """
    origin = get_origin(array=primary_beam, plot_origin=plot_origin)

    array_plotters.plot_array(
        array=primary_beam,
        origin=origin,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def get_origin(array, plot_origin):
    """Get the (y,x) origin of the ccd instrument if it going to be plotted.

    Parameters
    -----------
    array : instrument.array.scaled_array.ScaledArray
        The array from which the origin is extracted.
    plot_origin : True
        If true, the origin of the data's coordinate system is returned.
    """
    if plot_origin:
        return array.origin
    else:
        return None



import logging
import numpy as np
import scipy.signal
from skimage.transform import resize, rescale

from autolens import exc
from autolens.array import grids
from autolens.data.instrument import abstract_data
from autolens.array.mapping_util import grid_mapping_util

from autolens.array.util import array_util, grid_util
from autolens.array.mapping_util import array_mapping_util
from autolens.array import scaled_array

logger = logging.getLogger(__name__)


class InterferometerData(abstract_data.AbstractData):
    def __init__(
        self,
        image,
        pixel_scale,
        psf,
        noise_map,
        visibilities,
        visibilities_noise_map,
        uv_wavelengths,
        primary_beam,
        exposure_time_map=None,
    ):

        super(InterferometerData, self).__init__(
            image=image,
            pixel_scale=pixel_scale,
            psf=psf,
            noise_map=noise_map,
            exposure_time_map=exposure_time_map,
            origin=(0.0, 0.0),
        )

        self.visibilities = visibilities
        self.visibilities_magnitudes = np.sqrt(
            np.square(visibilities[:, 0]) + np.square(visibilities[:, 1])
        )
        self.visibilities_noise_map = visibilities_noise_map
        self.uv_wavelengths = uv_wavelengths
        self.primary_beam = primary_beam

    def new_interferometer_data_with_resized_arrays(
        self, new_shape, new_centre_pixels=None, new_centre_arcsec=None
    ):

        image = self.resize_scaled_array(
            scaled_array=self.image,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        noise_map = self.resize_scaled_array(
            scaled_array=self.noise_map,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        exposure_time_map = self.resize_scaled_array(
            scaled_array=self.exposure_time_map,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        return InterferometerData(
            image=image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=noise_map,
            exposure_time_map=exposure_time_map,
            visibilities=self.visibilities,
            visibilities_noise_map=self.visibilities_noise_map,
            uv_wavelengths=self.uv_wavelengths,
            primary_beam=self.primary_beam,
        )

    def new_interferometer_data_with_resized_psf(self, new_shape):
        psf = self.resize_scaled_array(scaled_array=self.psf, new_shape=new_shape)
        return InterferometerData(
            image=self.image,
            pixel_scale=self.pixel_scale,
            psf=psf,
            noise_map=self.noise_map,
            exposure_time_map=self.exposure_time_map,
            visibilities=self.visibilities,
            visibilities_noise_map=self.visibilities_noise_map,
            uv_wavelengths=self.uv_wavelengths,
            primary_beam=self.primary_beam,
        )

    def new_interferometer_data_with_resized_primary_beam(self, new_shape):
        primary_beam = self.resize_scaled_array(
            scaled_array=self.primary_beam, new_shape=new_shape
        )
        return InterferometerData(
            image=self.image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=self.noise_map,
            exposure_time_map=self.exposure_time_map,
            visibilities=self.visibilities,
            visibilities_noise_map=self.visibilities_noise_map,
            uv_wavelengths=self.uv_wavelengths,
            primary_beam=primary_beam,
        )

    def new_interferometer_data_converted_from_electrons(self):

        image = self.array_from_counts_to_electrons_per_second(array=self.image)
        noise_map = self.array_from_counts_to_electrons_per_second(array=self.noise_map)

        return InterferometerData(
            image=image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=noise_map,
            exposure_time_map=self.exposure_time_map,
            visibilities=self.visibilities,
            visibilities_noise_map=self.visibilities_noise_map,
            uv_wavelengths=self.uv_wavelengths,
            primary_beam=self.primary_beam,
        )

    def new_interferometer_data_converted_from_adus(self, gain):

        image = self.array_from_adus_to_electrons_per_second(
            array=self.image, gain=gain
        )
        noise_map = self.array_from_adus_to_electrons_per_second(
            array=self.noise_map, gain=gain
        )

        return InterferometerData(
            image=image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=noise_map,
            exposure_time_map=self.exposure_time_map,
            visibilities=self.visibilities,
            visibilities_noise_map=self.visibilities_noise_map,
            uv_wavelengths=self.uv_wavelengths,
            primary_beam=self.primary_beam,
        )


class NoiseMap(abstract_data.AbstractNoiseMap):

    pass


class PrimaryBeam(scaled_array.ScaledSquarePixelArray):

    # noinspection PyUnusedLocal
    def __init__(self, array, pixel_scale, renormalize=False, **kwargs):
        """
        Class storing a 2D Point Spread Function (PrimaryBeam), including its blurring kernel.

        Parameters
        ----------
        array : ndarray
            The 2d PrimaryBeam blurring kernel.
        renormalize : bool
            Renormalize the PrimaryBeam such that he sum of kernel values total 1.0?
        """

        # noinspection PyArgumentList
        super().__init__(array=array, pixel_scale=pixel_scale)
        if renormalize:
            self[:, :] = np.divide(self, np.sum(self))

    @classmethod
    def from_gaussian(
        cls, shape, pixel_scale, sigma, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0
    ):
        """Simulate the primary beam as an elliptical Gaussian profile."""
        from autolens.model.profiles.light_profiles import EllipticalGaussian

        gaussian = EllipticalGaussian(
            centre=centre, axis_ratio=axis_ratio, phi=phi, intensity=1.0, sigma=sigma
        )

        from autolens.array import grids

        grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=shape, pixel_scale=pixel_scale, sub_grid_size=1
        )
        gaussian = gaussian.profile_image_from_grid(
            grid=grid, return_in_2d=True, return_binned=True
        )

        return PrimaryBeam(array=gaussian, pixel_scale=pixel_scale, renormalize=True)

    @classmethod
    def from_fits_renormalized(cls, file_path, hdu, pixel_scale):
        """Loads a primary beam from fits and renormalizes it

        Parameters
        ----------
        pixel_scale
        file_path: String
            The path to the file containing the PrimaryBeam
        hdu : int
            The HDU the PrimaryBeam is stored in the .fits file.

        Returns
        -------
        primary_beam: PrimaryBeam
            A renormalized PrimaryBeam instance
        """
        primary_beam = PrimaryBeam.from_fits_with_scale(file_path, hdu, pixel_scale)
        primary_beam[:, :] = np.divide(primary_beam, np.sum(primary_beam))
        return primary_beam

    @classmethod
    def from_fits_with_scale(cls, file_path, hdu, pixel_scale):
        """
        Loads the PrimaryBeam from a .fits file.

        Parameters
        ----------
        pixel_scale
        file_path: String
            The path to the file containing the PrimaryBeam
        hdu : int
            The HDU the PrimaryBeam is stored in the .fits file.
        """
        return cls(
            array=array_util.numpy_array_2d_from_fits(file_path, hdu),
            pixel_scale=pixel_scale,
        )

    def new_primary_beam_with_rescaled_odd_dimensioned_array(
        self, rescale_factor, renormalize=True
    ):

        primary_beam_rescaled = rescale(
            self,
            rescale_factor,
            anti_aliasing=False,
            mode="constant",
            multichannel=False,
        )

        if (
            primary_beam_rescaled.shape[0] % 2 == 0
            and primary_beam_rescaled.shape[1] % 2 == 0
        ):
            primary_beam_rescaled = resize(
                primary_beam_rescaled,
                output_shape=(
                    primary_beam_rescaled.shape[0] + 1,
                    primary_beam_rescaled.shape[1] + 1,
                ),
                anti_aliasing=False,
                mode="constant",
            )
        elif (
            primary_beam_rescaled.shape[0] % 2 == 0
            and primary_beam_rescaled.shape[1] % 2 != 0
        ):
            primary_beam_rescaled = resize(
                primary_beam_rescaled,
                output_shape=(
                    primary_beam_rescaled.shape[0] + 1,
                    primary_beam_rescaled.shape[1],
                ),
                anti_aliasing=False,
                mode="constant",
            )
        elif (
            primary_beam_rescaled.shape[0] % 2 != 0
            and primary_beam_rescaled.shape[1] % 2 == 0
        ):
            primary_beam_rescaled = resize(
                primary_beam_rescaled,
                output_shape=(
                    primary_beam_rescaled.shape[0],
                    primary_beam_rescaled.shape[1] + 1,
                ),
                anti_aliasing=False,
                mode="constant",
            )

        pixel_scale_factors = (
            self.shape[0] / primary_beam_rescaled.shape[0],
            self.shape[1] / primary_beam_rescaled.shape[1],
        )
        pixel_scale = (
            self.pixel_scale * pixel_scale_factors[0],
            self.pixel_scale * pixel_scale_factors[1],
        )
        return PrimaryBeam(
            array=primary_beam_rescaled,
            pixel_scale=np.max(pixel_scale),
            renormalize=renormalize,
        )

    def new_primary_beam_with_renormalized_array(self):
        """Renormalize the PrimaryBeam such that its data_vector values sum to unity."""
        return PrimaryBeam(array=self, pixel_scale=self.pixel_scale, renormalize=True)

    def convolve(self, array_2d):
        """
        Convolve an array with this PrimaryBeam

        Parameters
        ----------
        image : ndarray
            An array representing the image the PrimaryBeam is convolved with.

        Returns
        -------
        convolved_image : ndarray
            An array representing the image after convolution.

        Raises
        ------
        KernelException if either PrimaryBeam primary_beam dimension is odd
        """
        if self.shape[0] % 2 == 0 or self.shape[1] % 2 == 0:
            raise exc.ConvolutionException("PrimaryBeam Kernel must be odd")

        return scipy.signal.convolve2d(array_2d, self, mode="same")


class SimulatedInterferometerData(InterferometerData):
    def __init__(
        self,
        image,
        pixel_scale,
        psf,
        noise_map,
        visibilities,
        visibilities_noise_map,
        uv_wavelengths,
        primary_beam,
        visibilities_noise_map_realization,
        exposure_time_map=None,
        **kwargs
    ):

        super(SimulatedInterferometerData, self).__init__(
            image=image,
            pixel_scale=pixel_scale,
            psf=psf,
            noise_map=noise_map,
            visibilities=visibilities,
            visibilities_noise_map=visibilities_noise_map,
            uv_wavelengths=uv_wavelengths,
            primary_beam=primary_beam,
            exposure_time_map=exposure_time_map,
        )

        self.visibilities_noise_map_realization = visibilities_noise_map_realization

    @classmethod
    def from_deflections_galaxies_and_exposure_arrays(
        cls,
        deflections,
        pixel_scale,
        galaxies,
        exposure_time,
        transformer,
        primary_beam=None,
        exposure_time_map=None,
        background_sky_level=0.0,
        background_sky_map=None,
        noise_sigma=None,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):

        shape = (deflections.shape[0], deflections.shape[1])

        grid_1d = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
            shape=shape, pixel_scale=pixel_scale
        )

        deflections_1d = grid_mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_grid_size(
            sub_grid_2d=deflections,
            mask=np.full(shape=shape, fill_value=False),
            sub_grid_size=1,
        )

        deflected_grid_1d = grid_1d - deflections_1d

        image_2d = sum(
            map(
                lambda g: g.profile_image_from_grid(
                    grid=deflected_grid_1d, return_in_2d=True, return_binned=False
                ),
                galaxies,
            )
        )

        return cls.from_image_and_exposure_arrays(
            image=image_2d,
            pixel_scale=pixel_scale,
            exposure_time=exposure_time,
            exposure_time_map=exposure_time_map,
            background_sky_level=background_sky_level,
            background_sky_map=background_sky_map,
            transformer=transformer,
            primary_beam=primary_beam,
            noise_sigma=noise_sigma,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

    @classmethod
    def from_tracer_grid_and_exposure_arrays(
        cls,
        tracer,
        grid,
        pixel_scale,
        exposure_time,
        transformer,
        primary_beam=None,
        exposure_time_map=None,
        background_sky_level=0.0,
        background_sky_map=None,
        noise_sigma=None,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and Interferometer read-out).
        pixel_scale: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """

        image_plane_image_2d = tracer.profile_image_from_grid(
            grid=grid, return_in_2d=True, return_binned=True
        )

        return cls.from_image_and_exposure_arrays(
            image=image_plane_image_2d,
            pixel_scale=pixel_scale,
            exposure_time=exposure_time,
            exposure_time_map=exposure_time_map,
            background_sky_level=background_sky_level,
            background_sky_map=background_sky_map,
            transformer=transformer,
            primary_beam=primary_beam,
            noise_sigma=noise_sigma,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
        )

    @classmethod
    def from_image_and_exposure_arrays(
        cls,
        image,
        pixel_scale,
        exposure_time,
        transformer,
        primary_beam=None,
        exposure_time_map=None,
        background_sky_level=0.0,
        background_sky_map=None,
        noise_sigma=None,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and Interferometer read-out).
        pixel_scale: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """

        if exposure_time_map is None:

            exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
                value=exposure_time, shape=image.shape, pixel_scale=pixel_scale
            )

        if background_sky_map is None:

            background_sky_map = scaled_array.ScaledSquarePixelArray.single_value(
                value=background_sky_level, shape=image.shape, pixel_scale=pixel_scale
            )

        image += background_sky_map
        image_1d = array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=image,
            mask=np.full(fill_value=False, shape=image.shape),
            sub_grid_size=1,
        )

        visibilities = transformer.visibilities_from_image_1d(image_1d=image_1d)

        if noise_sigma is not None:
            visibilities_noise_map_realization = gaussian_noise_map_from_shape_and_sigma(
                shape=visibilities.shape, sigma=noise_sigma, noise_seed=noise_seed
            )
            visibilities = visibilities + visibilities_noise_map_realization
            visibilities_noise_map = NoiseMap.single_value(
                value=noise_sigma, shape=visibilities.shape, pixel_scale=pixel_scale
            )
        else:
            visibilities_noise_map = NoiseMap.single_value(
                value=noise_if_add_noise_false,
                shape=visibilities.shape,
                pixel_scale=pixel_scale,
            )
            visibilities_noise_map_realization = None

        if np.isnan(visibilities_noise_map).any():
            raise exc.DataException(
                "The noise-map has NaN values in it. This suggests your exposure time and / or"
                "background sky levels are too low, creating signal counts at or close to 0.0."
            )

        image -= background_sky_map

        return SimulatedInterferometerData(
            image=image,
            pixel_scale=pixel_scale,
            noise_map=None,
            psf=None,
            visibilities=visibilities,
            visibilities_noise_map=visibilities_noise_map,
            uv_wavelengths=transformer.uv_wavelengths,
            primary_beam=primary_beam,
            visibilities_noise_map_realization=visibilities_noise_map_realization,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            noise_realization=visibilities_noise_map_realization,
        )

    def __array_finalize__(self, obj):
        if isinstance(obj, SimulatedInterferometerData):
            try:
                self.image = obj.image
                self.pixel_scale = obj.pixel_scale
                self.psf = obj.psf
                self.noise_map = obj.noise_map
                self.background_noise_map = obj.background_noise_map
                self.poisson_noise_map = obj.poisson_noise_map
                self.exposure_time_map = obj.exposure_time_map
                self.background_sky_map = obj.background_sky_map
                self.background_noise_realization = obj.background_noise_realization
                self.poisson_noise_realization = obj.poisson_noise_realization
                self.origin = obj.origin
            except AttributeError:
                logger.debug(
                    "Original object in Interferometer.__array_finalize__ missing one or more attributes"
                )


def gaussian_noise_map_from_shape_and_sigma(shape, sigma, noise_seed=-1):
    """Generate a two-dimensional read noises-map, generating values from a Gaussian distribution with mean 0.0.

    Params
    ----------
    shape : (int, int)
        The (x,y) image_shape of the generated Gaussian noises map.
    read_noise : float
        Standard deviation of the 1D Gaussian that each noises value is drawn from
    seed : int
        The seed of the random number generator, used for the random noises maps.
    """
    if noise_seed == -1:
        # Use one seed, so all regions have identical column non-uniformity.
        noise_seed = np.random.randint(0, int(1e9))
    np.random.seed(noise_seed)
    read_noise_map = np.random.normal(loc=0.0, scale=sigma, size=shape)
    return read_noise_map


def load_interferometer_data_from_fits(
    image_path,
    pixel_scale,
    real_visibilities_path=None,
    imaginary_visibilities_path=None,
    visibilities_noise_map_path=None,
    u_wavelengths_path=None,
    v_wavelengths_path=None,
    image_hdu=0,
    resized_interferometer_shape=None,
    resized_interferometer_origin_pixels=None,
    resized_interferometer_origin_arcsec=None,
    psf_path=None,
    psf_hdu=0,
    resized_psf_shape=None,
    renormalize_psf=True,
    resized_primary_beam_shape=None,
    renormalize_primary_beam=True,
    noise_map_path=None,
    noise_map_hdu=0,
    convert_noise_map_from_weight_map=False,
    convert_noise_map_from_inverse_noise_map=False,
    exposure_time_map_path=None,
    exposure_time_map_hdu=0,
    exposure_time_map_from_single_value=None,
    real_visibilities_hdu=0,
    imaginary_visibilities_hdu=0,
    visibilities_noise_map_hdu=0,
    u_wavelengths_hdu=0,
    v_wavelengths_hdu=0,
    primary_beam_path=None,
    primary_beam_hdu=0,
    convert_from_electrons=False,
    gain=None,
    convert_from_adus=False,
):
    """Factory for loading the interferometer instrument from .fits files, as well as computing properties like the noise-map,
    exposure-time map, etc. from the interferometer-instrument.

    This factory also includes a number of routines for converting the interferometer-instrument from units not supported by PyAutoLens \
    (e.g. adus, electrons) to electrons per second.

    Parameters
    ----------
    lens_name
    image_path : str
        The path to the image .fits file containing the image (e.g. '/path/to/image.fits')
    pixel_scale : float
        The size of each pixel in arc seconds.
    image_hdu : int
        The hdu the image is contained in the .fits file specified by *image_path*.        
    image_hdu : int
        The hdu the image is contained in the .fits file that *image_path* points too.
    resized_interferometer_shape : (int, int) | None
        If input, the interferometer arrays that are image sized, e.g. the image, noise-maps) are resized to these dimensions.
    resized_interferometer_origin_pixels : (int, int) | None
        If the interferometer arrays are resized, this defines a new origin (in pixels) around which recentering occurs.
    resized_interferometer_origin_arcsec : (float, float) | None
        If the interferometer arrays are resized, this defines a new origin (in arc-seconds) around which recentering occurs.
    primary_beam_path : str
        The path to the primary_beam .fits file containing the primary_beam (e.g. '/path/to/primary_beam.fits')        
    primary_beam_hdu : int
        The hdu the primary_beam is contained in the .fits file specified by *primary_beam_path*.
    resized_primary_beam_shape : (int, int) | None
        If input, the primary_beam is resized to these dimensions.
    renormalize_psf : bool
        If True, the PrimaryBeam is renoralized such that all elements sum to 1.0.
    noise_map_path : str
        The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')        
    noise_map_hdu : int
        The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
    noise_map_from_image_and_background_noise_map : bool
        If True, the noise-map is computed from the observed image and background noise-map \
        (see NoiseMap.from_image_and_background_noise_map).
    convert_noise_map_from_weight_map : bool
        If True, the noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_noise_map_from_inverse_noise_map : bool
        If True, the noise-map loaded from the .fits file is converted from an inverse noise-map to a noise-map (see \
        *NoiseMap.from_inverse_noise_map).
    background_noise_map_path : str
        The path to the background_noise_map .fits file containing the background noise-map \ 
        (e.g. '/path/to/background_noise_map.fits')        
    background_noise_map_hdu : int
        The hdu the background_noise_map is contained in the .fits file specified by *background_noise_map_path*.
    convert_background_noise_map_from_weight_map : bool
        If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_background_noise_map_from_inverse_noise_map : bool
        If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    poisson_noise_map_path : str
        The path to the poisson_noise_map .fits file containing the Poisson noise-map \
         (e.g. '/path/to/poisson_noise_map.fits')        
    poisson_noise_map_hdu : int
        The hdu the poisson_noise_map is contained in the .fits file specified by *poisson_noise_map_path*.
    poisson_noise_map_from_image : bool
        If True, the Poisson noise-map is estimated using the image.
    convert_poisson_noise_map_from_weight_map : bool
        If True, the Poisson noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_poisson_noise_map_from_inverse_noise_map : bool
        If True, the Poisson noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    exposure_time_map_path : str
        The path to the exposure_time_map .fits file containing the exposure time map \ 
        (e.g. '/path/to/exposure_time_map.fits')        
    exposure_time_map_hdu : int
        The hdu the exposure_time_map is contained in the .fits file specified by *exposure_time_map_path*.
    exposure_time_map_from_single_value : float
        The exposure time of the interferometer imaging, which is used to compute the exposure-time map as a single value \
        (see *ExposureTimeMap.from_single_value*).
    exposure_time_map_from_inverse_noise_map : bool
        If True, the exposure-time map is computed from the background noise_map map \
        (see *ExposureTimeMap.from_background_noise_map*)
    background_sky_map_path : str
        The path to the background_sky_map .fits file containing the background sky map \
        (e.g. '/path/to/background_sky_map.fits').
    background_sky_map_hdu : int
        The hdu the background_sky_map is contained in the .fits file specified by *background_sky_map_path*.
    convert_from_electrons : bool
        If True, the input unblurred_image_1d are in units of electrons and all converted to electrons / second using the exposure \
        time map.
    gain : float
        The image gain, used for convert from ADUs.
    convert_from_adus : bool
        If True, the input unblurred_image_1d are in units of adus and all converted to electrons / second using the exposure \
        time map and gain.
    """

    image = abstract_data.load_image(
        image_path=image_path, image_hdu=image_hdu, pixel_scale=pixel_scale
    )

    psf = abstract_data.load_psf(
        psf_path=psf_path,
        psf_hdu=psf_hdu,
        pixel_scale=pixel_scale,
        renormalize=renormalize_psf,
    )

    noise_map = load_noise_map(
        noise_map_path=noise_map_path,
        noise_map_hdu=noise_map_hdu,
        pixel_scale=pixel_scale,
        convert_noise_map_from_weight_map=convert_noise_map_from_weight_map,
        convert_noise_map_from_inverse_noise_map=convert_noise_map_from_inverse_noise_map,
    )

    exposure_time_map = abstract_data.load_exposure_time_map(
        exposure_time_map_path=exposure_time_map_path,
        exposure_time_map_hdu=exposure_time_map_hdu,
        pixel_scale=pixel_scale,
        shape=image.shape,
        exposure_time=exposure_time_map_from_single_value,
        exposure_time_map_from_inverse_noise_map=False,
        inverse_noise_map=None,
    )

    real_visibilities = load_visibilities(
        visibilities_path=real_visibilities_path, visibilities_hdu=real_visibilities_hdu
    )
    imaginary_visibilities = load_visibilities(
        visibilities_path=imaginary_visibilities_path,
        visibilities_hdu=imaginary_visibilities_hdu,
    )

    visibilities = np.stack((real_visibilities, imaginary_visibilities), axis=-1)

    visibilities_noise_map = load_visibilities_noise_map(
        visibilities_noise_map_path=visibilities_noise_map_path,
        visibilities_noise_map_hdu=visibilities_noise_map_hdu,
    )
    u_wavelengths = load_visibilities(
        visibilities_path=u_wavelengths_path, visibilities_hdu=u_wavelengths_hdu
    )
    v_wavelengths = load_visibilities(
        visibilities_path=v_wavelengths_path, visibilities_hdu=v_wavelengths_hdu
    )

    uv_wavelengths = np.stack((u_wavelengths, v_wavelengths), axis=-1)

    primary_beam = load_primary_beam(
        primary_beam_path=primary_beam_path,
        primary_beam_hdu=primary_beam_hdu,
        pixel_scale=pixel_scale,
        renormalize=renormalize_primary_beam,
    )

    interferometer_data = InterferometerData(
        image=image,
        pixel_scale=pixel_scale,
        psf=psf,
        primary_beam=primary_beam,
        noise_map=noise_map,
        visibilities=visibilities,
        visibilities_noise_map=visibilities_noise_map,
        uv_wavelengths=uv_wavelengths,
        exposure_time_map=exposure_time_map,
    )

    if resized_interferometer_shape is not None:
        interferometer_data = interferometer_data.new_interferometer_data_with_resized_arrays(
            new_shape=resized_interferometer_shape,
            new_centre_pixels=resized_interferometer_origin_pixels,
            new_centre_arcsec=resized_interferometer_origin_arcsec,
        )

    if resized_psf_shape is not None:
        interferometer_data = interferometer_data.new_interferometer_data_with_resized_psf(
            new_shape=resized_psf_shape
        )

    if resized_primary_beam_shape is not None:
        interferometer_data = interferometer_data.new_interferometer_data_with_resized_primary_beam(
            new_shape=resized_primary_beam_shape
        )

    if convert_from_electrons:
        interferometer_data = (
            interferometer_data.new_interferometer_data_converted_from_electrons()
        )
    elif convert_from_adus:
        interferometer_data = interferometer_data.new_interferometer_data_converted_from_adus(
            gain=gain
        )

    return interferometer_data


def load_noise_map(
    noise_map_path,
    noise_map_hdu,
    pixel_scale,
    convert_noise_map_from_weight_map,
    convert_noise_map_from_inverse_noise_map,
):
    """Factory for loading the noise-map from a .fits file.

    This factory also includes a number of routines for converting the noise-map from from other units (e.g. \
    a weight map) or computing the noise-map from other unblurred_image_1d (e.g. the interferometer image and background noise-map).

    Parameters
    ----------
    noise_map_path : str
        The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')
    noise_map_hdu : int
        The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
    pixel_scale : float
        The size of each pixel in arc seconds.
    image : ndarray
        The image-image, which the noise-map can be calculated using.
    background_noise_map : ndarray
        The background noise-map, which the noise-map can be calculated using.
    exposure_time_map : ndarray
        The exposure-time map, which the noise-map can be calculated using.
    convert_noise_map_from_weight_map : bool
        If True, the noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_noise_map_from_inverse_noise_map : bool
        If True, the noise-map loaded from the .fits file is converted from an inverse noise-map to a noise-map (see \
        *NoiseMap.from_inverse_noise_map).
    background_noise_map_path : str
        The path and filename of the .fits image containing the background noise-map.
    background_noise_map_hdu : int
        The hdu the background noise-map is contained in the .fits file that *background_noise_map_path* points too.
    convert_background_noise_map_from_weight_map : bool
        If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_background_noise_map_from_inverse_noise_map : bool
        If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    noise_map_from_image_and_background_noise_map : bool
        If True, the noise-map is computed from the observed image and background noise-map \
        (see NoiseMap.from_image_and_background_noise_map).
    convert_from_electrons : bool
        If True, the input unblurred_image_1d are in units of electrons and all converted to electrons / second using the exposure \
        time map.
    gain : float
        The image gain, used for convert from ADUs.
    convert_from_adus : bool
        If True, the input unblurred_image_1d are in units of adus and all converted to electrons / second using the exposure \
        time map and gain.
    """
    noise_map_options = sum(
        [convert_noise_map_from_weight_map, convert_noise_map_from_inverse_noise_map]
    )

    if noise_map_options > 1:
        raise exc.DataException(
            "You have specified more than one method to load the noise_map map, e.g.:"
            "convert_noise_map_from_weight_map | "
            "convert_noise_map_from_inverse_noise_map |"
            "noise_map_from_image_and_background_noise_map"
        )

    if noise_map_options == 0 and noise_map_path is not None:
        return NoiseMap.from_fits_with_pixel_scale(
            file_path=noise_map_path, hdu=noise_map_hdu, pixel_scale=pixel_scale
        )
    elif convert_noise_map_from_weight_map and noise_map_path is not None:
        weight_map = scaled_array.Array.from_fits(
            file_path=noise_map_path, hdu=noise_map_hdu
        )
        return NoiseMap.from_weight_map(weight_map=weight_map, pixel_scale=pixel_scale)
    elif convert_noise_map_from_inverse_noise_map and noise_map_path is not None:
        inverse_noise_map = scaled_array.Array.from_fits(
            file_path=noise_map_path, hdu=noise_map_hdu
        )
        return NoiseMap.from_inverse_noise_map(
            inverse_noise_map=inverse_noise_map, pixel_scale=pixel_scale
        )
    else:
        raise exc.DataException(
            "A noise_map map was not loaded, specify a noise_map_path or option to compute a noise_map map."
        )


def load_visibilities(visibilities_path, visibilities_hdu):

    if visibilities_path is not None:
        return array_util.numpy_array_1d_from_fits(
            file_path=visibilities_path, hdu=visibilities_hdu
        )


def load_visibilities_noise_map(
    visibilities_noise_map_path, visibilities_noise_map_hdu
):
    if visibilities_noise_map_path is not None:
        return array_util.numpy_array_1d_from_fits(
            file_path=visibilities_noise_map_path, hdu=visibilities_noise_map_hdu
        )


def load_baselines(baselines_path, baselines_hdu):
    if baselines_path is not None:
        return array_util.numpy_array_1d_from_fits(
            file_path=baselines_path, hdu=baselines_hdu
        )


def load_primary_beam(
    primary_beam_path, primary_beam_hdu, pixel_scale, renormalize=False
):
    """Factory for loading the primary_beam from a .fits file.

    Parameters
    ----------
    primary_beam_path : str
        The path to the primary_beam .fits file containing the primary_beam (e.g. '/path/to/primary_beam.fits')
    primary_beam_hdu : int
        The hdu the primary_beam is contained in the .fits file specified by *primary_beam_path*.
    pixel_scale : float
        The size of each pixel in arc seconds.
    renormalize : bool
        If True, the PrimaryBeam is renoralized such that all elements sum to 1.0.
    """
    if renormalize:
        return PrimaryBeam.from_fits_renormalized(
            file_path=primary_beam_path, hdu=primary_beam_hdu, pixel_scale=pixel_scale
        )
    if not renormalize:
        return PrimaryBeam.from_fits_with_scale(
            file_path=primary_beam_path, hdu=primary_beam_hdu, pixel_scale=pixel_scale
        )


def output_interferometer_data_to_fits(
    interferometer_data,
    image_path,
    psf_path,
    noise_map_path=None,
    primary_beam_path=None,
    exposure_time_map_path=None,
    real_visibilities_path=None,
    imaginary_visibilities_path=None,
    visibilities_noise_map_path=None,
    u_wavelengths_path=None,
    v_wavelengths_path=None,
    overwrite=False,
):
    array_util.numpy_array_2d_to_fits(
        array_2d=interferometer_data.image, file_path=image_path, overwrite=overwrite
    )
    array_util.numpy_array_2d_to_fits(
        array_2d=interferometer_data.psf, file_path=psf_path, overwrite=overwrite
    )

    if primary_beam_path is not None:
        array_util.numpy_array_2d_to_fits(
            array_2d=interferometer_data.primary_beam,
            file_path=primary_beam_path,
            overwrite=overwrite,
        )

    if interferometer_data.noise_map is not None and noise_map_path is not None:
        array_util.numpy_array_2d_to_fits(
            array_2d=interferometer_data.noise_map,
            file_path=noise_map_path,
            overwrite=overwrite,
        )

    if (
        interferometer_data.exposure_time_map is not None
        and exposure_time_map_path is not None
    ):
        array_util.numpy_array_2d_to_fits(
            array_2d=interferometer_data.exposure_time_map,
            file_path=exposure_time_map_path,
            overwrite=overwrite,
        )

    if (
        interferometer_data.visibilities is not None
        and real_visibilities_path is not None
    ):
        array_util.numpy_array_1d_to_fits(
            array_1d=interferometer_data.visibilities[:, 0],
            file_path=real_visibilities_path,
            overwrite=overwrite,
        )

    if (
        interferometer_data.visibilities is not None
        and imaginary_visibilities_path is not None
    ):
        array_util.numpy_array_1d_to_fits(
            array_1d=interferometer_data.visibilities[:, 1],
            file_path=imaginary_visibilities_path,
            overwrite=overwrite,
        )

    if (
        interferometer_data.visibilities_noise_map is not None
        and visibilities_noise_map_path is not None
    ):
        array_util.numpy_array_1d_to_fits(
            array_1d=interferometer_data.visibilities_noise_map,
            file_path=visibilities_noise_map_path,
            overwrite=overwrite,
        )

    if (
        interferometer_data.uv_wavelengths is not None
        and u_wavelengths_path is not None
    ):
        array_util.numpy_array_1d_to_fits(
            array_1d=interferometer_data.uv_wavelengths[:, 0],
            file_path=u_wavelengths_path,
            overwrite=overwrite,
        )

    if (
        interferometer_data.uv_wavelengths is not None
        and v_wavelengths_path is not None
    ):
        array_util.numpy_array_1d_to_fits(
            array_1d=interferometer_data.uv_wavelengths[:, 1],
            file_path=v_wavelengths_path,
            overwrite=overwrite,
        )



from astropy import cosmology
from scipy.optimize import fsolve
from autolens.model.profiles import mass_profiles as mp
from astropy import constants
import numpy as np

from autolens.model import cosmology_util

cosmo = cosmology.FlatLambdaCDM(H0=70.0, Om0=0.3)

z_lens = 0.6
z_source = 2.5

C_v = constants.c.to("km / s")
G_crit = 4.302e-6  # kpc M_sun^-1 (km/s)^2
G_crit = constants.G.to("kpc km2 / (solMass s2)")

s_crit = (
    C_v
    * C_v
    * cosmo.angular_diameter_distance(z_source)
    / (
        4.0
        * np.pi
        * cosmo.angular_diameter_distance(z_lens)
        * cosmo.angular_diameter_distance_z1z2(z_lens, z_source)
    )
    / (1000.0 * G_crit)
)  # critical density at los halo

const = (C_v * C_v) / (4 * np.pi) / (1000.0 * G_crit)

print(const)

angs = cosmo.angular_diameter_distance(z_source) / (
    cosmo.angular_diameter_distance(z_lens)
    * cosmo.angular_diameter_distance_z1z2(z_lens, z_source)
)

s_crit_al = cosmology_util.critical_surface_density_between_redshifts_from_redshifts_and_cosmology(
    redshift_0=z_lens,
    redshift_1=z_source,
    cosmology=cosmo,
    unit_length="kpc",
    unit_mass="solMass",
)

print(s_crit)
print(s_crit_al)



from autolens import exc
from autolens.plotters import plotter_util

import matplotlib.pyplot as plt
import numpy as np


def plot_line(
    y,
    x,
    as_subplot=False,
    label=None,
    plot_axis_type="semilogy",
    vertical_lines=None,
    vertical_line_labels=None,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    plot_legend=False,
    title="Quantity vs Radius",
    ylabel="Quantity",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    legend_fontsize=12,
    output_path=None,
    output_format="show",
    output_filename="quantity_vs_radius",
):

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)
    plotter_util.set_title(title=title, titlesize=titlesize)

    if x is None:
        x = np.arange(len(y))

    plot_y_vs_x(y=y, x=x, plot_axis_type=plot_axis_type, label=label)

    set_xy_labels_and_ticksize(
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        ylabel=ylabel,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
    )

    plot_vertical_lines(
        vertical_lines=vertical_lines,
        vertical_line_labels=vertical_line_labels,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
    )

    set_legend(plot_legend=plot_legend, legend_fontsize=legend_fontsize)

    plotter_util.output_figure(
        array=None,
        as_subplot=as_subplot,
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plotter_util.close_figure(as_subplot=as_subplot)


def plot_y_vs_x(y, x, plot_axis_type, label):

    if plot_axis_type is "linear":
        plt.plot(x, y, label=label)
    elif plot_axis_type is "semilogy":
        plt.semilogy(x, y, label=label)
    elif plot_axis_type is "loglog":
        plt.loglog(x, y, label=label)
    else:
        raise exc.PlottingException(
            "The plot_axis_type supplied to the plotter is not a valid string (must be linear "
            "| semilogy | loglog)"
        )


def set_xy_labels_and_ticksize(
    units, kpc_per_arcsec, ylabel, xlabelsize, ylabelsize, xyticksize
):
    """Set the x and y labels of the figure, and set the fontsize of those labels.

    The x label is always the distance scale / radius, thus the x-label is either arc-seconds or kpc and depending \
    on the units the figure is plotted in.

    The ylabel is the physical quantity being plotted and is passed as an input parameter.

    Parameters
    -----------
    units : str
        The units of the y / x axis of the plots, in arc-seconds ('arcsec') or kiloparsecs ('kpc').
    kpc_per_arcsec : float
        The conversion factor between arc-seconds and kiloparsecs, required to plot the units in kpc.
    ylabel : str
        The y-label of the figure, which is the physical quantitiy being plotted.
    xlabelsize : int
        The fontsize of the x axes label.
    ylabelsize : int
        The fontsize of the y axes label.
    xyticksize : int
        The font size of the x and y ticks on the figure axes.
    """

    plt.ylabel(ylabel=ylabel, fontsize=ylabelsize)

    if units in "arcsec" or kpc_per_arcsec is None:

        plt.xlabel("x (arcsec)", fontsize=xlabelsize)

    elif units in "kpc":

        plt.xlabel("x (kpc)", fontsize=xlabelsize)

    else:
        raise exc.PlottingException(
            "The units supplied to the plotter are not a valid string (must be pixels | "
            "arcsec | kpc)"
        )

    plt.tick_params(labelsize=xyticksize)


def plot_vertical_lines(vertical_lines, vertical_line_labels, units, kpc_per_arcsec):

    if vertical_lines is [] or vertical_lines is None:
        return

    for vertical_line, vertical_line_label in zip(vertical_lines, vertical_line_labels):

        if units in "arcsec" or kpc_per_arcsec is None:
            x_value_plot = vertical_line
        elif units in "kpc":
            x_value_plot = vertical_line
        else:
            raise exc.PlottingException(
                "The units supplied to the plotter are not a valid string (must be pixels | "
                "arcsec | kpc)"
            )

        plt.axvline(x=x_value_plot, label=vertical_line_label, linestyle="--")


def set_legend(plot_legend, legend_fontsize):
    if plot_legend:
        plt.legend(fontsize=legend_fontsize)



import autofit as af
import autolens as al
from test.integration.tests import runner

test_type = "model_mapper"
test_name = "link_variable"
data_type = "lens_light_dev_vaucouleurs"
data_resolution = "LSST"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):
    class MMPhase(al.PhaseImaging):
        pass

    phase1 = MMPhase(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, light=al.light_profiles.EllipticalSersic)
        ),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 20
    phase1.optimizer.sampling_efficiency = 0.8

    class MMPhase2(al.PhaseImaging):
        def customize_priors(self, results):

            self.galaxies.lens = results.from_phase("phase_1").variable.lens

    phase2 = MMPhase2(
        phase_name="phase_2",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, light=al.light_profiles.EllipticalSersic)
        ),
        optimizer_class=optimizer_class,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 20
    phase2.optimizer.sampling_efficiency = 0.8

    return al.PipelineImaging(name, phase1, phase2)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import autofit as af

# In this pipeline, we'll demonstrate the use of an inversion pixel limit - which allows us to put an upper limit on
# the number of pixels an inversion uses in a pipeline. In this example, we will perform an initial analysis which is
# restricted to just 100 pixels, for super fast run-times, and then refine the model in a second phase at the input
# resolution.

# Whilst the inversion pixel limit can be manually specified in the pipeline, in this example we will make the limit
# an input parameter of the pipeline. This means we can run the pipeline with different pixel limits for different
# runners.

# We will also use phase tagging to ensure phases which use a pixel limit have a tag in their path, so it is clear
# what settings a phases has when it uses this feature.

# We'll perform a basic analysis which fits a lensed source galaxy using an inversion where
# the lens's light is omitted. This pipeline uses two phases:

# Phase 1:

# Description: initialize the lens mass model and source light profile using an inversion with a pixel limit.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: VoronoiMagnification
# Previous Pipelines: None
# Prior Passing: None
# Notes: Uses an inversion pixel limit of 100

# Phase 2:

# Description: Fits the lens and source model without an inversion pixel limit
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: VoronoiMagnification
# Previous Pipelines: None
# Prior Passing: Lens mass (variable -> phase 1), source light (variable -> phase 1)
# Notes: No pixel limit


def make_pipeline(phase_folders=None, inversion_pixel_limit=100):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_feature__inversion_pixel_limit"

    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings()

    # When a phase is passed a inversion_pixel_limit, a settings tag is automatically generated and added to the
    # phase path to make it clear what binning up was used. The settings tag, phase name and phase paths are shown for 3
    # example inversion pixel limits:

    # inversion_pixel_limit=50 -> phase_path=phase_name/settings_pix_lim_50
    # inversion_pixel_limit=80 -> phase_path=phase_name/settings_pix_lim_80

    # If the inversion pixel limit is None, the tag is an empty string, thus not changing the settings tag:

    # inversion_pixel_limit=None -> phase_path=phase_name/settings
    # inversion_pixel_limit=1 -> phase_path=phase_name/settings

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/pipeline_name/pipeline_tag/phase_name/phase_tag//'

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    # As there is no lens light component, we can use an annular mask throughout this pipeline which removes the
    # central regions of the image.

    def mask_function(image):
        return al.Mask.circular_annular(
            shape=image.shape,
            pixel_scale=image.pixel_scale,
            inner_radius_arcsec=0.2,
            outer_radius_arcsec=3.3,
        )

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Limit the number of source pixels used by the inversion.

    class LensSourceX1Phase(al.PhaseImaging):
        def customize_priors(self, results):

            self.galaxies.lens.mass.centre_0 = af.GaussianPrior(mean=0.0, sigma=0.1)

            self.galaxies.lens.mass.centre_1 = af.GaussianPrior(mean=0.0, sigma=0.1)

    phase1 = LensSourceX1Phase(
        phase_name="phase_1__x1_source",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=al.pixelizations.VoronoiMagnification,
                regularization=al.regularization.Constant,
            ),
        ),
        mask_function=mask_function,
        inversion_pixel_limit=inversion_pixel_limit,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # In phase 2, we will fit the lens galaxy's mass and two source galaxies, where we:

    # 1) Omit the inversion pixel limit, thus performing the modeling at a high source plane resolution if necessary.

    class LensSourceX2Phase(al.PhaseImaging):
        def customize_priors(self, results):

            self.galaxies.lens = results.from_phase(
                "phase_1__x1_source"
            ).variable.galaxies.lens

            self.galaxies.source = results.from_phase(
                "phase_1__x1_source"
            ).variable.galaxies.source

    phase2 = LensSourceX2Phase(
        phase_name="phase_2__x2_source",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=al.mass_profiles.ExternalShear,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=al.pixelizations.VoronoiMagnification,
                regularization=al.regularization.Constant,
            ),
        ),
        mask_function=mask_function,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.3

    return al.PipelineImaging(pipeline_name, phase1, phase2)



import autofit as af
import autolens as al

import os

# In this example, we will load the image of a strong lens from a .fits file and plot it using the
# function autolens.instrument.plotters.array_plotters.plot_array. We will customize the appearance of this figure to
# highlight the features of the image. For more generical plotting tools (e.g. changing the figure size, axis units,
# outputting the image to the hard-disk, etc.) checkout the example in 'workspace/plotting/examples/arrays/array.py'.

# We will use the image of slacs1430+4105.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# We have included the .fits instrument required for this example in the directory
# 'workspace/output/data/example/slacs1430+4105/'.

# First, lets setup the path to the .fits file of the image.
data_type = "example"
data_name = "slacs1430+4105"

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/example/slacs1430+4105/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)
image_path = data_path + "image.fits"

# Now, lets load this array as a hyper array. A hyper array is an ordinary NumPy array, but it also includes a pixel
# scale which allows us to convert the axes of the array to arc-second coordinates.
image = al.ScaledSquarePixelArray.from_fits_with_pixel_scale(
    file_path=image_path, hdu=0, pixel_scale=0.03
)

# We can now use an array plotter to plot the image. Lets first plot it using the default PyAutoLens settings.
al.array_plotters.plot_array(array=image, title="SLACS1430+4105 Image")

# For a lens like SLACS1430+4105, the lens galaxy's light outshines the background source, making it appear faint.
# we can use a symmetric logarithmic colorbar normalization to better reveal the source galaxy (due to negative values
# in the image, we cannot use a logirithmic colorbar normalization).
al.array_plotters.plot_array(
    array=image,
    title="SLACS1430+4105 Image",
    norm="symmetric_log",
    linthresh=0.05,
    linscale=0.02,
)

# Alternatively, we can use the default linear colorbar normalization and customize the limits over which the colormap
# spans its dynamic range.
al.array_plotters.plot_array(
    array=image, title="SLACS1430+4105 Image", norm="linear", norm_min=0.0, norm_max=0.3
)

psf_path = workspace_path + "/data/example/" + data_name + "/psf.fits"
noise_map_path = workspace_path + "/data/example/" + data_name + "/noise_map.fits"

ccd_data = al.load_ccd_data_from_fits(
    image_path=data_path,
    psf_path=psf_path,
    noise_map_path=noise_map_path,
    pixel_scale=0.03,
)

# These plotters can be customized using the exact same functions as above.
al.ccd_plotters.plot_noise_map(
    ccd_data=ccd_data, title="SLACS1430+4105 Noise-Map", norm="log"
)

# Of course, as we've seen in many other examples, a sub-plot of the ccd instrument can be plotted. This can also take the
# customization inputs above, but it should be noted that the options are applied to all images, and thus will most
# likely degrade a number of the sub-plot images.
al.ccd_plotters.plot_ccd_subplot(
    ccd_data=ccd_data, norm="symmetric_log", linthresh=0.05, linscale=0.02
)



import sys
import math
import numpy as np
import matplotlib.pyplot as plt

sys.path.append("../")

from paper_plots import density_plot_tools
from autolens.analysis import galaxy
from autolens.model.profiles import mass_profiles


ltm_skip = 0
center_skip = 0

image_dir = (
    "/gpfs/data_vector/pdtw24/PL_Data/SL03_2/"
)  # Dir of Object to make evidence tables from

image_name = "SLACSJ0252+0039"
image_name = "SLACSJ1250+0523"
# image_name = 'SLACSJ1430+4105'

slacs = density_plot_tools.SLACS(image_dir, image_name)

# pipeline_folder = 'Pipeline_Total'
# phase_folder = 'PL_Phase_8'

# pipeline_folder = 'Pipeline_LMDM'
# phase_folder = 'PL_Phase_7'

pipeline_folder = "Pipeline_LTM2"
phase_folder = "PL_5_LMDM_LTM_R_MC"

sersic_bulge = mass_profiles.EllipticalSersicMass(
    centre=(0.02, 0.008),
    axis_ratio=0.87,
    phi=71.1,
    intensity=0.09458,
    effective_radius=1.0153,
    sersic_index=4.57468,
    mass_to_light_ratio=1.55265,
)

exponential_halo = mass_profiles.EllipticalSersicMassRadialGradient(
    centre=(0.043, 0.036),
    axis_ratio=0.7861,
    phi=161.34,
    intensity=0.02475,
    effective_radius=2.1389,
    mass_to_light_ratio=1.2556,
    mass_to_light_gradient=-0.2229,
)

dark_matter_halo = mass_profiles.SphericalNFW(
    kappa_s=0.1035, scale_radius=30.0 * slacs.arcsec_per_kpc
)

lens_galaxy = galaxy.Galaxy(
    redshift=slacs.redshift,
    mass_profiles=[sersic_bulge, exponential_halo, dark_matter_halo],
)
source_galaxy = galaxy.Galaxy(redshift=slacs.source_redshift)

number_bins = 30
radius_kpc = 30

radii = list(np.linspace(5e-3, radius_kpc * slacs.arcsec_per_kpc, number_bins + 1))

radii_plot = []
bulge_densities = []
halo_densities = []
dark_densities = []

for r in range(number_bins):

    annuli_area = math.pi * radii[r + 1] ** 2 - math.pi * radii[r] ** 2

    densities = (
        lens_galaxy.dimensionless_mass_within_circles_individual(radii[r + 1])
        - lens_galaxy.dimensionless_mass_within_circles_individual(radii[r])
    ) / annuli_area

    bulge_densities.append(densities[0])
    halo_densities.append(densities[1])
    dark_densities.append(densities[2])

    radii_plot.append(((radii[r + 1] + radii[r]) / 2.0) * slacs.kpc_per_arcsec)

print(bulge_densities)

plt.semilogy(radii_plot, bulge_densities, color="r", label="Sersic Bulge")
plt.semilogy(
    radii_plot, halo_densities, color="g", label="EllipticalExponentialMass Halo"
)
plt.semilogy(radii_plot, dark_densities, color="k", label="Dark Matter Halo")

plt.legend()
plt.show()

# slacs.weighted_densities_vs_radii(radius_kpc=30.0, weight_cut=1e-3, number_bins=20)

# slacs.plot_density(image_name=image_name, labels=['Sersic', 'Exponential', 'NFWSph'])



import autofit as af
import autolens as al


# In this pipeline, we'll perform a basic analysis which fits two source galaxies using a parametric light profile and
# a lens galaxy where its light is not present in the image, using two phases:

# Phase 1:

# Description: initialize the lens mass model and source light profile using x1 source.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: None

# Phase 2:

# Description: Fit the lens mass model and source light profile using x2 sources.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Galaxy 1 - Light: EllipticalSersic
# Source Galaxy 2 - Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: Lens mass (variable -> phase 1), Source Galaxy 1 Light (variable -> phase 1)
# Notes: None


def make_pipeline(
    include_shear=True,
    phase_folders=None,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    signal_to_noise_limit=None,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    pixel_scale_interpolation_grid=None,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_lens_sie__source_x2_sersic"

    pipeline_tag = al.pipeline_tagging.pipeline_tag_from_pipeline_settings()

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/pipeline_name/pipeline_tag/phase_name/phase_tag/'

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    ### SETUP SHEAR ###

    # If the pipeline should include shear, add this class below so that it enters the phase.

    # After this pipeline this shear class is passed to all subsequent pipelines, such that the shear is either
    # included or omitted throughout the entire pipeline.

    if include_shear:
        shear = al.mass_profiles.ExternalShear
    else:
        shear = None

    # As there is no lens light component, we can use an annular mask throughout this pipeline which removes the
    # central regions of the image.

    def mask_function(image):
        return al.Mask.circular_annular(
            shape=image.shape,
            pixel_scale=image.pixel_scale,
            inner_radius_arcsec=0.2,
            outer_radius_arcsec=3.3,
        )

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy (y,x) centre such that we assume the image is centred around the lens galaxy.

    class LensSourceX1Phase(al.PhaseImaging):
        def customize_priors(self, results):

            self.galaxies.lens.mass.centre_0 = af.GaussianPrior(mean=0.0, sigma=0.1)
            self.galaxies.lens.mass.centre_1 = af.GaussianPrior(mean=0.0, sigma=0.1)

    phase1 = LensSourceX1Phase(
        phase_name="phase_1__lens_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                mass=al.mass_profiles.EllipticalIsothermal,
                shear=shear,
            ),
            source_0=al.GalaxyModel(
                redshift=redshift_source, light=al.light_profiles.EllipticalSersic
            ),
        ),
        mask_function=mask_function,
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        optimizer_class=af.MultiNest,
    )

    # You'll see these lines throughout all of the example pipelines. They are used to make MultiNest sample the \
    # non-linear parameter space faster (if you haven't already, checkout 'tutorial_7_multinest_black_magic' in
    # 'howtolens/chapter_2_lens_modeling'.

    # Fitting the lens galaxy and source galaxy from uninitialized priors often risks MultiNest getting stuck in a
    # local maxima, especially for the image in this example which actually has two source galaxies. Therefore, whilst
    # I will continue to use constant efficiency mode to ensure fast run time, I've upped the number of live points
    # and decreased the sampling efficiency from the usual values to ensure the non-linear search is robust.

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # In phase 2, we will fit the lens galaxy's mass and two source galaxies, where we:

    # 1) Initialize the priors on the lens galaxy using the results of phase 1.
    # 2) Initialize the priors on the first source galaxy using the results of phase 1.

    phase2 = al.PhaseImaging(
        phase_name="phase_2__lens_sie__source_x2_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=redshift_lens,
                mass=phase1.result.variable.galaxies.lens.mass,
                shear=phase1.result.variable.galaxies.lens.shear,
            ),
            source_0=al.GalaxyModel(
                redshift=redshift_source,
                light=phase1.result.variable.galaxies.source_0.light,
            ),
            source_1=al.GalaxyModel(
                redshift=redshift_source, light=al.light_profiles.EllipticalSersic
            ),
        ),
        mask_function=mask_function,
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.3

    return al.PipelineImaging(pipeline_name, phase1, phase2)



import os

import autofit as af
import autolens as al

test_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))


def pixel_scale_from_data_resolution(data_resolution):
    """Determine the pixel scale from a instrument resolution type based on real observations.

    These options are representative of LSST, Euclid, HST, over-sampled HST and Adaptive Optics image.

    Parameters
    ----------
    data_resolution : str
        A string giving the resolution of the desired instrument type (LSST | Euclid | HST | HST_Up | AO).
    """
    if data_resolution == "LSST":
        return 0.2
    elif data_resolution == "Euclid":
        return 0.1
    elif data_resolution == "HST":
        return 0.05
    elif data_resolution == "HST_Up":
        return 0.03
    elif data_resolution == "AO":
        return 0.01
    else:
        raise ValueError(
            "An invalid instrument resolution was entered - ", data_resolution
        )


def shape_from_data_resolution(data_resolution):
    """Determine the shape of an image from a instrument resolution type based on real observations.

    These options are representative of LSST, Euclid, HST, over-sampled HST and Adaptive Optics image.

    Parameters
    ----------
    data_resolution : str
        A string giving the resolution of the desired instrument type (LSST | Euclid | HST | HST_Up | AO).
    """
    if data_resolution == "LSST":
        return (100, 100)
    elif data_resolution == "Euclid":
        return (150, 150)
    elif data_resolution == "HST":
        return (250, 250)
    elif data_resolution == "HST_Up":
        return (320, 320)
    elif data_resolution == "AO":
        return (750, 750)
    else:
        raise ValueError("An invalid instrument-type was entered - ", data_resolution)


def data_resolution_from_pixel_scale(pixel_scale):
    if pixel_scale == 0.2:
        return "LSST"
    elif pixel_scale == 0.1:
        return "Euclid"
    elif pixel_scale == 0.05:
        return "HST"
    elif pixel_scale == 0.03:
        return "HST_Up"
    elif pixel_scale == 0.01:
        return "AO"
    else:
        raise ValueError("An invalid pixel-scale was entered - ", pixel_scale)


def load_test_ccd_data(data_type, data_resolution, psf_shape=(11, 11), lens_name=None):
    pixel_scale = pixel_scale_from_data_resolution(data_resolution=data_resolution)

    data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=test_path, folder_names=["data", data_type, data_resolution]
    )

    return al.load_ccd_data_from_fits(
        image_path=data_path + "/image.fits",
        psf_path=data_path + "/psf.fits",
        noise_map_path=data_path + "/noise_map.fits",
        pixel_scale=pixel_scale,
        resized_psf_shape=psf_shape,
        lens_name=lens_name,
    )



import autofit as af
import autolens as al

import os

# This tool allows one to mask a bespoke mask for a given image of a strong lens, which can then be loaded before a
# pipeline is run and passed to that pipeline so as to become the default masked used by a phase (if a mask
# function is not passed to that phase).

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# The 'data name' is the name of the data folder and 'data_name' the folder the mask is stored in, e.g,
# the mask will be output as '/workspace/data/data_type/data_name/mask.fits'.
data_type = "example"
data_name = "lens_sie__source_sersic"

# Create the path where the mask will be output, which in this case is
# '/workspace/data/example/lens_light_and_x1_source/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

# If you use this tool for your own instrument, you *must* double check this pixel scale is correct!
pixel_scale = 0.1

# First, load the CCD imaging data, so that the mask can be plotted over the strong lens image.
image = al.load_image(
    image_path=data_path + "image.fits", image_hdu=0, pixel_scale=pixel_scale
)

# Now, create a mask for this instrument, using the mask function's we're used to. I'll use a circular-annular mask here,
# but I've commented over options you might want to use (feel free to experiment!)

mask = al.Mask.circular_annular(
    shape=image.shape,
    pixel_scale=image.pixel_scale,
    inner_radius_arcsec=0.5,
    outer_radius_arcsec=2.5,
    centre=(0.0, 0.0),
)

# Now lets plot the image and mask, so we can check that the mask includes the regions of the image we want.
al.data_plotters.plot_image(image=image, mask=mask)

# Now we're happy with the mask, lets output it to the data folder of the lens, so that we can load it from a .fits
# file in our pipelines!
al.output_mask_to_fits(mask=mask, mask_path=data_path + "mask.fits", overwrite=True)



import autofit as af
import autolens as al

# In this pipeline, we'll perform a basic analysis which fits a source galaxy using an inversion and a
# lens galaxy where its light is not included and fitted, using two phases:

# Phase 1) Fit the lens galaxy's mass (SIE) and source galaxy's light (Sersic).

# Phase 2) Fit the lens galaxy's mass (SIE) and source galaxy's light using an inversion, where the SIE mass model
#          priors are initialized from phase 1. This new SIE mass model will be used to refine the inversion's
#          pixelization and regularization parameters.

# Phase 3) Fit the lens galaxy's mass (SIE) and source galaxy's light using another inversion, which has had the
#           pixelization and regularization refined from phase 2.


def make_pipeline(phase_folders=None):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline__inversion"

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/'
    phase_folders.append(pipeline_name)

    # This is the same phase 1 as the complex source pipeline, which we saw gave a good fit to the overall
    # structure of the lensed source and provided an accurate lens mass model.

    phase1 = al.PhaseImaging(
        phase_name="phase_1__lens_sie__source_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=al.mass_profiles.EllipticalIsothermal
            ),
            source=al.GalaxyModel(
                redshift=1.0, light=al.light_profiles.EllipticalSersic
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.sampling_efficiency = 0.3
    phase1.optimizer.const_efficiency_mode = True

    # Now, in phase 2, lets use the lens mass model to fit the source with an inversion.

    class InversionPhase(al.PhaseImaging):
        def customize_priors(self, results):

            # We can customize the inversion's priors like we do our light and mass profiles.

            self.galaxies.lens = results.from_phase(
                "phase_1__lens_sie__source_sersic"
            ).variable.galaxies.lens

            self.galaxies.source.pixelization.shape_0 = af.UniformPrior(
                lower_limit=20.0, upper_limit=40.0
            )

            self.galaxies.source.pixelization.shape_1 = af.UniformPrior(
                lower_limit=20.0, upper_limit=40.0
            )

            # The expected value of the regularization coefficient depends on the details of the data reduction and
            # source galaxy. A broad log-uniform prior is thus an appropriate way to sample the large range of
            # possible values.

            self.galaxies.source.regularization.coefficient = af.LogUniformPrior(
                lower_limit=1.0e-6, upper_limit=10000.0
            )

    phase2 = InversionPhase(
        phase_name="phase_2__source_inversion_initialize",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=al.mass_profiles.EllipticalIsothermal
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=al.pixelizations.VoronoiMagnification,
                regularization=al.regularization.Constant,
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.sampling_efficiency = 0.3
    phase2.optimizer.const_efficiency_mode = True

    # This function 'extends' phase 2, with an additional 'inversion phase'. An inversion phase is a standard phase
    # which automatically uses the best-fit model of the phase it is extending (in this example, phase 2) to refine the
    # inversion by optimizing only the pixelization & regularization parameters.

    # The results of this inversion phase are accessible in their own attribute of the results, called 'inversion, as
    # we'll see in phase 3.

    phase2 = phase2.extend_with_inversion_phase()

    # Now, in phase 3, lets use the refined source inversion to fit the lens mass model again.

    class InversionPhase(al.PhaseImaging):
        def customize_priors(self, results):
            # We can customize the inversion's priors like we do our light and mass profiles.

            self.galaxies.lens = results.from_phase(
                "phase_2__source_inversion_initialize"
            ).variable.galaxies.lens

            # Note the use of the 'inversion' result attribute below, which uses the extended inversion phase results.

            self.galaxies.source = results.from_phase(
                "phase_2__source_inversion_initialize"
            ).inversion.constant.galaxies.source

    phase3 = InversionPhase(
        phase_name="phase_3__lens_sie__source_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5, mass=al.mass_profiles.EllipticalIsothermal
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=al.pixelizations.VoronoiMagnification,
                regularization=al.regularization.Constant,
            ),
        ),
        optimizer_class=af.MultiNest,
    )

    phase3.optimizer.sampling_efficiency = 0.3
    phase3.optimizer.const_efficiency_mode = True

    return al.PipelineImaging(pipeline_name, phase1, phase2, phase3)



import autofit as af
import autolens as al
from test.integration.tests import runner

test_type = "lens_only"
test_name = "lens_light__hyper"
data_type = "lens_bulge_disk"
data_resolution = "LSST"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):
    phase1 = al.PhaseImaging(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, light=al.light_profiles.EllipticalSersic)
        ),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    phase1 = phase1.extend_with_multiple_hyper_phases(hyper_galaxy=True)

    class HyperLensPlanePhase(al.PhaseImaging):
        def customize_priors(self, results):

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

    phase2 = HyperLensPlanePhase(
        phase_name="phase_2",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=phase1.result.variable.galaxies.lens.light,
                hyper_galaxy=al.HyperGalaxy,
            )
        ),
        optimizer_class=optimizer_class,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 40
    phase2.optimizer.sampling_efficiency = 0.8

    return al.PipelineImaging(name, phase1, phase2)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



from autolens import decorator_util

import numpy as np


class Transformer(object):
    def __init__(self, uv_wavelengths, grid_radians, preload_transform=True):

        self.uv_wavelengths = uv_wavelengths
        self.grid_radians = grid_radians

        self.total_visibilities = uv_wavelengths.shape[0]
        self.total_image_pixels = grid_radians.shape[0]

        self.preload_transform = preload_transform

        if preload_transform:

            self.preload_real_transforms = self.preload_real_transforms(
                grid_radians=grid_radians,
                uv_wavelengths=uv_wavelengths,
                total_image_pixels=self.total_image_pixels,
            )

            self.preload_imaginary_transforms = self.preload_imaginary_transforms(
                grid_radians=grid_radians,
                uv_wavelengths=uv_wavelengths,
                total_image_pixels=self.total_image_pixels,
            )

    def real_visibilities_from_image_1d(self, image_1d):

        if self.preload_transform:

            return self.real_visibilities_via_preload_jit(
                image_1d=image_1d,
                preloaded_reals=self.preload_real_transforms,
                total_visibilities=self.total_visibilities,
                total_image_pixels=self.total_image_pixels,
            )

        else:

            return self.real_visibilities_jit(
                image_1d=image_1d,
                grid_radians=self.grid_radians,
                uv_wavelengths=self.uv_wavelengths,
                total_visibilities=self.total_visibilities,
                total_image_pixels=self.total_image_pixels,
            )

    @staticmethod
    @decorator_util.jit()
    def preload_real_transforms(grid_radians, uv_wavelengths, total_image_pixels):

        preloaded_real_transforms = np.zeros(
            shape=(total_image_pixels, uv_wavelengths.shape[0])
        )

        for i in range(total_image_pixels):
            for j in range(uv_wavelengths.shape[0]):
                preloaded_real_transforms[i, j] += np.cos(
                    -2.0
                    * np.pi
                    * (
                        grid_radians[i, 1] * uv_wavelengths[j, 0]
                        - grid_radians[i, 0] * uv_wavelengths[j, 1]
                    )
                )

        return preloaded_real_transforms

    @staticmethod
    @decorator_util.jit()
    def real_visibilities_via_preload_jit(
        image_1d, preloaded_reals, total_visibilities, total_image_pixels
    ):

        real_visibilities = np.zeros(shape=(total_visibilities))

        for i in range(total_image_pixels):
            for j in range(total_visibilities):
                real_visibilities[j] += image_1d[i] * preloaded_reals[i, j]

        return real_visibilities

    @staticmethod
    @decorator_util.jit()
    def real_visibilities_jit(
        image_1d, grid_radians, uv_wavelengths, total_visibilities, total_image_pixels
    ):

        real_visibilities = np.zeros(shape=(total_visibilities))

        for i in range(total_image_pixels):
            for j in range(total_visibilities):
                real_visibilities[j] += image_1d[i] * np.cos(
                    -2.0
                    * np.pi
                    * (
                        grid_radians[i, 1] * uv_wavelengths[j, 0]
                        - grid_radians[i, 0] * uv_wavelengths[j, 1]
                    )
                )

        return real_visibilities

    def imaginary_visibilities_from_image_1d(self, image_1d):

        if self.preload_transform:

            return self.imaginary_visibilities_via_preload_jit(
                image_1d=image_1d,
                preloaded_imaginarys=self.preload_imaginary_transforms,
                total_visibilities=self.total_visibilities,
                total_image_pixels=self.total_image_pixels,
            )

        else:

            return self.imaginary_visibilities_jit(
                image_1d=image_1d,
                grid_radians=self.grid_radians,
                uv_wavelengths=self.uv_wavelengths,
                total_visibilities=self.total_visibilities,
                total_image_pixels=self.total_image_pixels,
            )

    @staticmethod
    @decorator_util.jit()
    def preload_imaginary_transforms(grid_radians, uv_wavelengths, total_image_pixels):

        preloaded_imaginary_transforms = np.zeros(
            shape=(total_image_pixels, uv_wavelengths.shape[0])
        )

        for i in range(total_image_pixels):
            for j in range(uv_wavelengths.shape[0]):
                preloaded_imaginary_transforms[i, j] += np.sin(
                    -2.0
                    * np.pi
                    * (
                        grid_radians[i, 1] * uv_wavelengths[j, 0]
                        - grid_radians[i, 0] * uv_wavelengths[j, 1]
                    )
                )

        return preloaded_imaginary_transforms

    @staticmethod
    @decorator_util.jit()
    def imaginary_visibilities_via_preload_jit(
        image_1d, preloaded_imaginarys, total_visibilities, total_image_pixels
    ):

        imaginary_visibilities = np.zeros(shape=(total_visibilities))

        for i in range(total_image_pixels):
            for j in range(total_visibilities):
                imaginary_visibilities[j] += image_1d[i] * preloaded_imaginarys[i, j]

        return imaginary_visibilities

    @staticmethod
    @decorator_util.jit()
    def imaginary_visibilities_jit(
        image_1d, grid_radians, uv_wavelengths, total_visibilities, total_image_pixels
    ):

        imaginary_visibilities = np.zeros(shape=(total_visibilities))

        for i in range(total_image_pixels):
            for j in range(total_visibilities):
                imaginary_visibilities[j] += image_1d[i] * np.sin(
                    -2.0
                    * np.pi
                    * (
                        grid_radians[i, 1] * uv_wavelengths[j, 0]
                        - grid_radians[i, 0] * uv_wavelengths[j, 1]
                    )
                )

        return imaginary_visibilities

    def visibilities_from_image_1d(self, image_1d):

        real_visibilities = self.real_visibilities_from_image_1d(image_1d=image_1d)
        imaginary_visibilities = self.imaginary_visibilities_from_image_1d(
            image_1d=image_1d
        )

        return np.stack((real_visibilities, imaginary_visibilities), axis=-1)
